/*
 *	xmisc.c			X-Tuitate miscellous routines for X11
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/wait.h>
#include <X11/X.h>
#include <X11/xpm.h>

#include "config.h"

#include "all.h"
#include "misc.h"
#include "resource.h"
#include "xmisc.h"

void xpmError(int status, char *s)
{
    char *msg;

    switch( status ) {
    case XpmColorError:
	fprintf(stderr, "Warning: Pixmap color allocation pertially failed\n");
    case XpmSuccess:
	break;
    case XpmNoMemory:
	msg = "No enough memory to create pixmap";
	goto putmsg;
    case XpmColorFailed:
	msg = "Pixmap color allocation failed";
    putmsg:
	if( s != NULL && *s )
	    fprintf(stderr, "%s: ", s);
	fprintf(stderr, "%s\n", msg);
	break;
    }
}

void CreatePixmapFromData(display, d, data,
			  pixmap_return, shapemask_return, attributes)
    Display *display;
    Drawable d;
    char **data;
    Pixmap *pixmap_return;
    Pixmap *shapemask_return;
    XpmAttributes *attributes;
{
    int status;

    status = XpmCreatePixmapFromData(display, d, data, pixmap_return,
				     shapemask_return, attributes);
    if( status != XpmSuccess ) {
	xpmError(status, "XpmCreatePixmapFromData");
	if( status != XpmColorError )
	    exit(1);
    }
}

void create_pixmap_from_data(MyPixmap *mypixmap, char **data)
{
    int status;
    XpmAttributes attr;

    attr.valuemask = 0;

    status = XpmCreatePixmapFromData(xdata.display, win.id, data,
				     &mypixmap->pixmap, &mypixmap->shapemask,
				     &attr);
    if( status != XpmSuccess ) {
	fprintf(stderr, "%s: XpmCreatePixmapFromData(): %s\n",
		myname, XpmGetErrorString(status));
	exit(1);
    }
    mypixmap->width	= attr.width;
    mypixmap->height	= attr.height;
}

#if 0
Pixmap ReadFileToPixmap(char *filename)
{
    int status;
    Pixmap xpm;

    if( filename[0] == '-' && !filename[1] )
	filename = NULL;

    status = XpmReadFileToPixmap(xdata.display, xdata.root.id, filename,
				 &xpm, None, None);

    switch( status ) {
    case XpmColorError:
	fprintf(stderr, "Warning: Pixmap color allocation pertially failed\n");
    case XpmSuccess:
	break;
    case XpmOpenFailed:
    case XpmFileInvalid:
	perror(filename);
	exit(1);
    case XpmNoMemory:
	fprintf(stderr, "No enough memory to create pixmap.\n");
	exit(1);
    case XpmColorFailed:
	fprintf(stderr, "Pixmap color allocation failed\n");
	exit(1);
    }
    return xpm;
}

#else
static void fp_to_fd(FILE *fp, int fd)
{
    int len;
    char buf[1024];

    while( (len = fread(buf, 1, sizeof(buf), fp)) ) {
	if( ferror(fp) ) {
	    perror("fread");
	    exit(1);
	}
	if( write(fd, buf, len) < len ) {
	    perror("write");
	    exit(1);
	}
    }
}

static int wait_subprocess(pid_t pid, const char * name)
{
    int status;

    if( !pid )
	return 0;
    if( waitpid(pid, &status, WNOHANG) )
	if( WIFEXITED(status) ) {
	    status = WEXITSTATUS(status);
	    if( status ) {
		fprintf(stderr, "Child process (%s) exited with code %d.\n",
			name, status);
		return 1;
	    }
	    return 0;
	}
	else if( WIFSIGNALED(status) ) {
	    fprintf(stderr, "Chiled process (%s) was signaled %d%s\n",
		    name, WTERMSIG(status),
		    WCOREDUMP(status) ? " (core dumped)" : "");
	    return 1;
	}

    kill(pid, SIGTERM);
    waitpid(pid, &status, 0);
    return 1;
}

Pixmap ReadFileToPixmap(char *filename)
{
    int status;
    Pixmap xpm;

    if( filename[0] == '-' && !filename[1] ) {
	status = XpmReadFileToPixmap(xdata.display, xdata.root.id, NULL,
				     &xpm, None, None);
    }
    else {
	char *converter;
	size_t len;
	FILE *fp;
	int ch;
	unsigned char *ptr;
	unsigned char magic[9];
	pid_t child1 = 0;
	pid_t child2 = 0;
	pid_t child3 = 0;
	pid_t child4 = 0;
	pid_t child5 = 0;

	converter = getResourceStringNoError(".converter", ".Converter");
	if( Debugging('C') )
	    fprintf(stderr, "Converter program: %s\n", converter);

	if( (fp = fopen(filename, "r")) == NULL ) {
	    status = XpmOpenFailed;
	    goto error;
	}

	len = 0;

	if( (ch = fgetc(fp)) == EOF ) {
	    status = XpmFileInvalid;
	    goto error;
	}
	magic[0] = ch;
	len = 1;
#if defined(GZIP_PATH) || defined(UNCOMPRESS_PATH)
	if( Debugging('C') )
	    fprintf(stderr, "First byte: %o\n", ch);
	if( ch == 037 ) {
	    if( (ch = fgetc(fp)) == EOF ) {
		status = XpmFileInvalid;
		goto error;
	    }
	    magic[1] = ch;
	    len = 2;
	    if( Debugging('C') )
		fprintf(stderr, "Second byte: %o\n", ch);
	    if( ch == 0235
#ifdef GZIP_PATH
		|| ch == 0213
#endif
		) {
		int pfd1[2], pfd2[2];

		if( Debugging('C') )
		    fprintf(stderr, "%s: gzip'ed or compress'ed file.\n",
			    filename);

		if( pipe(pfd1) ) goto pipe_error;
		if( !(child1 = fork()) ) {
		    close(pfd1[0]);
		    if( write(pfd1[1], magic, 2) < 2 ) {
			perror(" Pipe write");
			exit(1);
		    }
		    fp_to_fd(fp, pfd1[1]);
		    while( fgetc(fp) != EOF ) ;
		    exit(0);
		}
		if( child1 < 0 ) goto fork_error;
		if( Debugging('C') )
		    fprintf(stderr, "fork #1 succeeded\n");
		close(pfd1[1]);
		fclose(fp);
		if( pipe(pfd2) ) goto pipe_error;
		if( !(child2 = fork()) ) {
		    close(pfd2[0]);
		    if( dup2(pfd1[0], 0) < 0 || dup2(pfd2[1], 1) < 0 ) {
			perror("dup2");
			exit(6);
		    }
#ifdef GZIP_PATH
		    if( Debugging('C') )
			fprintf(stderr, "Execute: %s\n", GZIP_PATH);
		    execl(GZIP_PATH, "gzip", "-dc", NULL);
#else
		    execl(UNCOMPRESS_PATH, "uncompress", "-c", NULL );
#endif
		    perror("exec");
		    exit(8);
		}
		if( child2 < 0 ) goto fork_error;
		if( Debugging('C') )
		    fprintf(stderr, "fork #2 succeeded\n");
		close(pfd2[1]);
		if( (fp = fdopen(pfd2[0], "r")) == NULL ) {
		    perror("fdopen");
		    goto subp_error;		    
		}
		len = 0;
	    }
	}
	if( converter ) {
	    ptr = "/* XPM */";
	    if( len && memcmp(magic, ptr, len) )
		goto convert;
	    ptr += len;
	    while( *ptr ) {
		if( (ch = fgetc(fp)) == EOF )
		    goto convert;
		magic[len++] = ch;
		if( ch != *ptr )
		    goto convert;
		ptr++;
	    }
	}
	while( len ) {
	    if( ungetc(magic[len - 1], fp) == EOF ) {
		int pfd[2];

		if( pipe(pfd) ) goto pipe_error;
		if( !(child3 = fork()) ) {
		    close(pfd[0]);
		    if( write(pfd[1], magic, len) < len ) {
			perror("Pipe write");
			exit(1);
		    }
		    fp_to_fd(fp, pfd[1]);
		    while( fgetc(fp) != EOF ) ;
		    exit(0);
		}
		if( child3 < 0 ) goto fork_error;
		if( Debugging('C') )
		    fprintf(stderr, "fork #3 succeeded\n");
		fclose(fp);
		close(pfd[1]);
		if( (fp = fdopen(pfd[0], "r")) == NULL ) {
		    perror("fdopen");
		    goto subp_error;
		}
		break;
	    }
	    len--;
	}
	goto read_xpm;

    convert:
	{
	    int pfd1[2], pfd2[2];

	    if( pipe(pfd1) ) goto pipe_error;
	    if( !(child4 = fork()) ) {
		close(pfd1[0]);
		if( write(pfd1[1], magic, len) < len ) {
		    perror("Pipe write");
		    exit(1);
		}
		fp_to_fd(fp, pfd1[1]);
		while( fgetc(fp) != EOF ) ;
		exit(0);
	    }
	    if( child4 < 0 ) {

		int status;

	    fork_error:
		perror("Can't spawn subprocess");
	    subp_error:
		if( child1 ) {
		    kill(child1, SIGTERM); waitpid(child1, &status, 0);
		}
		if( child2 ) {
		    kill(child2, SIGTERM); waitpid(child2, &status, 0);
		}
		if( child3 ) {
		    kill(child3, SIGTERM); waitpid(child3, &status, 0);
		}
		if( child4 ) {
		    kill(child4, SIGTERM); waitpid(child4, &status, 0);
		}
		if( child5 ) {
		    kill(child5, SIGTERM); waitpid(child5, &status, 0);
		}
		exit(1);
	    }
	    if( Debugging('C') )
		fprintf(stderr, "fork #4 succeeded\n");
	    close(pfd1[1]);
	    fclose(fp);
	    if( pipe(pfd2) ) {

	    pipe_error:
		perror("Can't create pipe");
		goto subp_error;
	    }
	    if( !(child5 = fork()) ) {
		close(pfd2[0]);
		if( dup2(pfd1[0], 0) < 0 || dup2(pfd2[1], 1) < 0 ) {
		    perror("dup2");
		    exit(1);
		}
		shexec(converter);
		perror("shexec");
		exit(9);
	    }
	    if( child5 < 0 ) goto fork_error;
	    if( Debugging('C') )
		fprintf(stderr, "fork #5 succeeded\n");
	    close(pfd2[1]);
	    if( (fp = fdopen(pfd2[0], "r")) == NULL ) {
		perror("fdopen");
		goto subp_error;
	    }
	}
    read_xpm:
	{
	    FILE save_stdin = *stdin;
	    int status;

	    *stdin = *fp;	/* XXX danger ? */
	    status = XpmReadFileToPixmap(xdata.display, xdata.root.id,
					 NULL, &xpm, NULL, NULL);
	    *fp = *stdin;
	    while( fgetc(fp) != EOF ) ;
	    fclose(fp);
	    *stdin = save_stdin;

	    if( wait_subprocess(child1, "bridge 1") ||
		wait_subprocess(child2, "decompressor") ||
		wait_subprocess(child3, "bridge 2") ||
		wait_subprocess(child4, "converter") ||
		wait_subprocess(child5, "bridge 3") )
		exit(1);
	}
#else
	status = XpmReadFileToPixmap(xdata.display, xdata.root.id,
				     filename, &xpm, NULL, NULL);
#endif	
    }

error:
    switch( status ) {
    case XpmColorError:
	fprintf(stderr, "Warning: Pixmap color allocation pertially failed\n");
    case XpmSuccess:
	break;
    case XpmOpenFailed:
	fprintf(stderr, "%s: Open failed\n", filename);
	exit(1);
    case XpmFileInvalid:
	fprintf(stderr, "%s: Invalid file\n", filename);
	perror(filename);
	exit(1);
    case XpmNoMemory:
	fprintf(stderr, "No enough memory to create pixmap.\n");
	exit(1);
    case XpmColorFailed:
	fprintf(stderr, "Pixmap color allocation failed\n");
	exit(1);
    }
    return xpm;
}
#endif

void ding(void)
{
    XBell(xdata.display, 0);
    XFlush(xdata.display);
}

static unsigned long alloc_color(const char *typeout,
				 const char *name, const char *class,
				 unsigned long fallback)
{
    const char * color_name;
    XColor color;

    if( (color_name = getResourceStringNoError(name, class)) == NULL ) {
	fprintf(stderr, "%s: Warning: %s color not specified",
		myname, typeout);
	return fallback;
    }
    if( !XParseColor(xdata.display, xdata.cmap, color_name, &color) ) {
	fprintf(stderr, "%s: Warning: cannot look up %s color `%s'\n",
		myname, typeout, color_name);
	return fallback;
    }
    if( !XAllocColor(xdata.display, xdata.cmap, &color) ) {
	fprintf(stderr, "%s: Warning: cannot allocate %s color `%s'\n",
		myname, typeout, color_name);
	return fallback;
    }
    return color.pixel;
}

unsigned long foregroundPixel(void)
{
    static unsigned long fg_pixel;
    static int fg_flag = False;

    if( !fg_flag ) {
	fg_pixel = alloc_color("foreground", ".foreground", ".Foreground",
			       WhitePixel(xdata.display, xdata.screen));
	fg_flag = True;
    }
    return fg_pixel;
}

unsigned long backgroundPixel(void)
{
    static unsigned long bg_pixel;
    static int bg_flag = False;

    if( !bg_flag ) {
	bg_pixel = alloc_color("background", ".background", ".Background",
			       BlackPixel(xdata.display, xdata.screen));
	bg_flag = True;
    }
    return bg_pixel;
}
